from flask import jsonify
from flask_jwt_extended import JWTManager
from datalake_catalog.app import app

jwt = JWTManager(app)


@jwt.user_lookup_loader
def user_lookup_callback(jwt_header, jwt_payload):
    return {
        "user": jwt_payload["sub"],
        "role": jwt_payload["role"] if "role" in jwt_payload else "",
    }


@jwt.expired_token_loader
def expired_token_callback(jwt_header, jwt_payload):
    app.logger.warning(f"Received an expired token")
    return jsonify(message="Unauthorized"), 401


@jwt.invalid_token_loader
def invalid_token_callback(reason):
    app.logger.warning(f"Received an invalid token. {reason}")
    return jsonify(message="Unauthorized"), 401


@jwt.unauthorized_loader
def unauthorized_callback(reason):
    return jsonify(message="Unauthorized"), 401



import json
from string import Formatter
import re
from pony import orm
from pony.orm.core import ObjectNotFound
from pony.flask import Pony
from datalake_catalog.app import app
from urllib.parse import urlparse
from pkg_resources import resource_stream

db = orm.Database()
Pony(app)
provider_scheme = {"aws": "s3", "gcp": "gs", "azure": "adls", "local": "file"}


def connect(db_string):  # pragma: no cover
    r = urlparse(db_string)
    if r.scheme == "sqlite":
        path = r.path[1:]  # strip the first /
        db.bind(provider="sqlite", filename=path, create_db=True)
    elif r.scheme == "mysql":
        host = f"{r.hostname}:{r.port}" if r.port is not None else r.hostname
        db.bind(provider="mysql", host=host, user=r.username, passwd=r.password, db=path[1:])
    elif r.scheme == "postgresql":
        host = f"{r.hostname}:{r.port}" if r.port is not None else r.hostname
        db.bind(
            provider="postgres",
            user=r.username,
            password=r.password,
            host=host,
            database=path[1:],
        )
    elif r.scheme == "oracle":
        db.bind(provider="oracle", user=r.username, password=r.password, dsn=r.hostname)
    elif r.scheme == "local":
        db.bind(provider="sqlite", filename=":memory:")
    else:
        raise ValueError(f"Unknown provider '{r.scheme}' in database connection string")
    _init_db()


def _init_db():
    db.generate_mapping(create_tables=True)

    with resource_stream("datalake_catalog", f"config/default.json") as f:
        default_config = json.load(f)

    with orm.db_session:
        for key, value in default_config.items():
            if Configuration.get(key=key) is None:
                Configuration(key=key, value=value)


class Catalog(db.Entity):
    key = orm.PrimaryKey(str)
    spec = orm.Required(orm.Json)

    domain = orm.Required(str)
    provider = orm.Required(str)
    feed = orm.Required(str)
    path_pattern = orm.Required(str)

    @property
    def path_regex(self):
        formatter = Formatter().parse(self.path_pattern)
        fields = []
        result = ""
        for (literal_text, field_name, format_spec, conversion) in formatter:
            result += re.escape(literal_text)
            if field_name is not None:
                if field_name in fields:  # backreference for existing name
                    result += f"(?P={field_name})"
                else:
                    fields.append(field_name)  # new name reference
                    result += f"(?P<{field_name}>.+)"
        return re.compile(f"{result}$")  # ensure only suffixes are matched


class Storage(db.Entity):
    key = orm.PrimaryKey(str)
    bucket = orm.Required(str)
    prefix = orm.Optional(str)

    def get_real_path(self, path):
        if self.prefix is not None and len(self.prefix) > 0:
            return f"{self.prefix}/{path}"
        return f"{path}"

    def build_uri(self, path):
        provider = Configuration["provider"].value
        real_path = self.get_real_path(path)
        scheme = provider_scheme[provider]
        return f"{scheme}://{self.bucket}/{real_path}"


class Configuration(db.Entity):
    key = orm.PrimaryKey(str)
    value = orm.Required(orm.Json)


def upsert_catalog(key, spec):
    domain = spec["domain"]
    provider = spec["provider"]
    feed = spec["feed"]

    # Preprocess the path pattern
    path_pattern = spec["storage"]["path"]["pattern"]
    path_format = Formatter().parse(path_pattern)
    path_pattern = ""
    for (literal_text, field_name, format_spec, conversion) in path_format:
        path_pattern += literal_text
        if field_name is not None:
            if field_name == "domain":
                path_pattern += domain
            elif field_name == "provider":
                path_pattern += provider
            elif field_name == "feed":
                path_pattern += feed
            else:
                path_pattern += "{" + field_name + "}"

    try:
        e = Catalog[key]
        e.spec = spec
        e.domain = domain
        e.provider = provider
        e.feed = feed
        e.path_pattern
    except ObjectNotFound:
        e = Catalog(
            key=key,
            spec=spec,
            domain=domain,
            provider=provider,
            feed=feed,
            path_pattern=path_pattern,
        )


def insert_storage(key, bucket, prefix=None):
    if prefix is not None:
        s = Storage(key=key, bucket=bucket, prefix=prefix)
    else:
        s = Storage(key=key, bucket=bucket)



import click
from datetime import timedelta
from flask_jwt_extended import create_access_token
from pony.flask import Pony
from datalake_catalog.app import app
from datalake_catalog.model import connect
import datalake_catalog.security
import datalake_catalog.api


app.config.from_object("datalake_catalog.settings.Default")
if not app.config.from_envvar("CATALOG_SETTINGS", silent=True):
    app.config.from_object("datalake_catalog.settings.Develop")

connect(app.config["DB_STRING"])


@click.group()
def cli():
    pass


@cli.command()
@click.argument("name")
@click.option("-e", "--expires", type=int, help="number of days before token expires")
@click.option(
    "-r",
    "--role",
    type=click.Choice(["author", "admin"]),
    help="the role associated with the name",
)
def create_api_key(name, expires, role):
    claims = {}
    if role is not None:
        claims["role"] = role
    if expires is None:
        expiry = False
    else:
        expiry = timedelta(days=expires)
    with app.app_context():
        click.echo(
            create_access_token(
                identity=name,
                additional_claims=claims,
                expires_delta=expiry,
            )
        )



from os import urandom


class Default:
    DEBUG = False
    TESTING = False

    JWT_TOKEN_LOCATION = ["headers"]
    JWT_ALGORITHM = "HS256"
    JWT_DECODE_ALGORITHMS = "HS256"
    JWT_HEADER_NAME = "Authorization"
    JWT_HEADER_TYPE = "Bearer"
    JWT_ACCESS_TOKEN_EXPIRES = 60


class Develop(Default):
    ENV = "development"
    SECRET_KEY = b"\x97\x1f\x0f\xfb\xdaI\xcf&'\xa9\xe8z\xfa\x0f\xbcV\xefx\x16\x14\xe4\xd5\xf20\x92{J\xee\xf1\xae\x9aS"
    DB_STRING = "sqlite://localhost/catalog.sqlite"


class UnitTest(Default):
    TESTING = True
    SECRET_KEY = urandom(32)
    DB_STRING = "local://"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datalake_catalog', 'datalake_catalog.schemas']

package_data = \
{'': ['*'], 'datalake_catalog': ['config/*']}

install_requires = \
['Flask-JWT-Extended>=4.3.1,<5.0.0',
 'Flask>=2.0.2,<3.0.0',
 'click>=8.0.3,<9.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'jsonschema>=4.2.1,<5.0.0',
 'pony>=0.7.14,<0.8.0']

entry_points = \
{'console_scripts': ['catalog = datalake_catalog.main:cli']}

setup_kwargs = {
    'name': 'datalake-catalog',
    'version': '1.0.1',
    'description': 'Datalake Catalog',
    'long_description': None,
    'author': 'Didier SCHMITT',
    'author_email': 'dschmitt@equancy.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.equancy.cloud/equancy/data-technologies/datalake-catalog',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from flask import Flask, jsonify
from werkzeug.exceptions import HTTPException

app = Flask(__name__)


@app.errorhandler(HTTPException)
def handle_exception(error):
    if error.code >= 500:  # pragma: no cover
        app.logger.error(error)

    return jsonify(message=error.name), error.code



from flask import jsonify, request, abort
from flask_jwt_extended import jwt_required, current_user
from jsonschema.exceptions import ValidationError
from datalake_catalog.app import app
from datalake_catalog.model import (
    Catalog,
    upsert_catalog,
    Storage,
    insert_storage,
    Configuration,
)
from datalake_catalog.schemas import SchemaValidator
from string import Formatter

catalog_validator = SchemaValidator("catalog")
storage_validator = SchemaValidator("storage")
config_validator = SchemaValidator("configuration")


@app.errorhandler(ValidationError)
def handle_validation_error(error):
    return jsonify(message=f"at {error.json_path} : {error.message}"), 400


def check_role_author():
    if current_user["role"] not in ("admin", "author"):
        abort(403)


def check_role_admin():
    if current_user["role"] not in ("admin"):
        abort(403)


@app.get("/health")
def get_health():
    return jsonify(message="OK"), 200


@app.get("/catalog")
def get_catalog():
    if "full" in request.args:
        return jsonify({e.key: e.spec for e in Catalog.select()}), 200
    return jsonify([e.key for e in Catalog.select()]), 200


@app.get("/catalog/schema")
def get_catalog_schema():
    return jsonify(catalog_validator.schema), 200


@app.get("/catalog/entry/<entry_id>")
def get_catalog_entry(entry_id):
    e = Catalog.get(key=entry_id)
    if e is None:
        abort(404)
    r = e.spec
    r["_key"] = entry_id
    return jsonify(r), 200


@app.get("/catalog/storage/<entry_id>")
def get_catalog_storage(entry_id):
    e = Catalog.get(key=entry_id)
    if e is None:
        abort(404)

    entry_storage = e.spec["storage"]
    prefix_params = {}
    for param, value in request.args.items():
        if param in entry_storage["path"]["params"]:
            prefix_params[param] = value

    formatter = Formatter().parse(e.path_pattern)
    result = ""
    is_partial = False
    for (literal_text, field_name, format_spec, conversion) in formatter:
        result += literal_text
        if field_name is not None:
            if field_name in prefix_params:
                result += str(prefix_params[field_name])
            else:
                is_partial = True
                break

    return jsonify(prefix=result, is_partial=is_partial), 200


@app.get("/catalog/identify/<path:path>")
def get_catalog_identify(path):
    result = []
    for e in Catalog.select():
        m = e.path_regex.search(path)
        if m is not None:
            result.append({"entry": e.key, "params": m.groupdict()})
    if len(result) == 0:
        abort(404)
    return jsonify(result), 200


def validate_spec(spec):
    catalog_validator.validate(spec)


@app.put("/catalog/entry/<entry_id>")
@jwt_required()
def put_catalog_entry(entry_id):
    check_role_author()
    validate_spec(request.get_json())
    upsert_catalog(entry_id, request.get_json())
    app.logger.info(f"User '{current_user['user']}' changed the entry '{entry_id}'")
    return jsonify(message="OK"), 200


@app.post("/catalog/import")
@jwt_required()
def post_catalog_import():
    check_role_author()
    error_messages = {}
    for key, value in request.get_json().items():
        try:
            validate_spec(value)
        except ValidationError as error:
            error_messages[key] = f"at {error.json_path}: {error.message}"
    if len(error_messages.keys()) > 0:
        return (
            jsonify(message="Catalog validation failed", failures=error_messages),
            400,
        )

    if "truncate" in request.args:
        Catalog.select().delete(bulk=True)
        app.logger.info(f"User '{current_user['user']}' truncated the entries")
    for key, value in request.get_json().items():
        upsert_catalog(key, value)
        app.logger.info(f"User '{current_user['user']}' changed the entry '{key}'")
    return jsonify(message="OK"), 200


@app.get("/storage")
def get_storages():
    return (
        jsonify({s.key: {"bucket": s.bucket, "prefix": s.prefix} for s in Storage.select()}),
        200,
    )


@app.put("/storage")
@jwt_required()
def put_storage():
    check_role_admin()
    storage_validator.validate(request.get_json())
    Storage.select().delete(bulk=True)

    for key, value in request.get_json().items():
        insert_storage(key, value["bucket"], value["prefix"] if "prefix" in value else None)
    app.logger.info(f"User '{current_user['user']}' updated the Storage configuration")
    return jsonify(message="OK"), 200


@app.get("/storage/<store_id>/<path:path>")
def get_storage_path(store_id, path):
    s = Storage.get(key=store_id)
    if s is None:
        abort(404)
    real_path = s.get_real_path(path)
    uri = s.build_uri(path)
    return jsonify(bucket=s.bucket, path=real_path, uri=uri), 200


@app.get("/configuration")
def get_configuration():
    return jsonify({c.key: c.value for c in Configuration.select()}), 200


@app.put("/configuration")
@jwt_required()
def put_configuration():
    check_role_admin()
    config_validator.validate(request.get_json())
    for key, value in request.get_json().items():
        c = Configuration[key]
        c.value = value
    app.logger.info(f"User '{current_user['user']}' updated the global configuration")
    return jsonify(message="OK"), 200


@app.get("/configuration/<key>")
def get_configuration_entry(key):
    c = Configuration.get(key=key)
    if c is None:
        abort(404)
    return jsonify(c.value), 200


@app.put("/configuration/<key>")
@jwt_required()
def put_configuration_entry(key):
    check_role_admin()
    global_schema = config_validator.schema
    if key not in global_schema["properties"]:
        abort(404)

    local_schema = global_schema["properties"][key]
    local_schema["$schema"] = "http://json-schema.org/draft-07/schema#"
    SchemaValidator(local_schema).validate(request.get_json())

    c = Configuration.get(key=key)
    c.value = request.get_json()
    return jsonify(message="OK"), 200



import json
from pkg_resources import resource_stream
from jsonschema import Draft7Validator


class SchemaValidator:
    def __init__(self, schema):
        if isinstance(schema, dict):
            self._schema = schema
        else:
            with resource_stream("datalake_catalog", f"schemas/{schema}.json") as f:
                self._schema = json.load(f)
        Draft7Validator.check_schema(self._schema)
        self._validator = Draft7Validator(self._schema)

    @property
    def schema(self):
        return self._schema

    def validate(self, data):
        self._validator.validate(data)
